//! Classification of outliers
//!
//! WARNING: There's no formal/mathematical definition of what an outlier actually is. Therefore,
//! all outlier classifiers are *subjective*, however some classifiers that have become *de facto*
//! standard are provided here.

pub mod tukey;
